/*-------------------------------------------------------------------------------------
 Uses Olav Kallhovd master library for HX711 24-Bit Analog-to-Digital Converter for Weigh Scales
 Peter Balcombe, March 2021 
 Tested with  : Arduino Nano, two 10Kg load cells/HX711 A/Ds & 16x2 LCD + I2C backpack.
-----------------------------------,--------------------------------------------------
 This is a sketch which uses four HX711 modules to measure Centre of Gravity (CoG) and weight
 of a model, displaying vulus on an 16x2 I2C LCD display.
 Settling time (number of samples) and data filtering can be adjusted in the HX711 config.h file
 Sketch will initially calibrate on 1st use, using the Serial port to input known weights.
 Calibration values are stored in EEPROM, so thereafter the calibraion routine is bypassed.
 Subsequent Tare or Recalibration can be done via the Serial port by entering 't' or 'r'.
 Variables d1 (LE to front load point) & d2 (front/rear load point separation must be changed to
 reflect the CoG mechanical arrangement in use.
 Load cells arranged as #1 & #3 at Front, #2 & #4 at rear. Left/Right doesn't matter as summed.
 Note that the LCD module I2C address may need to be changed to suit the device in use
 (Use an I2C Sniffer sketch to detect ehe LCD address if no display present) 
*/

#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <HX711_ADC.h>
#include <EEPROM.h>

//pins:
const int HX711_dout_1 = 4; //mcu > HX711 no 1 dout pin
const int HX711_sck_1 = 5; //mcu > HX711 no 1 sck pin
const int HX711_dout_2 = 6; //mcu > HX711 no 2 dout pin
const int HX711_sck_2 = 7; //mcu > HX711 no 2 sck pin
const int HX711_dout_3 = 8; //mcu > HX711 no 1 dout pin
const int HX711_sck_3 = 9; //mcu > HX711 no 1 sck pin
const int HX711_dout_4 = 10; //mcu > HX711 no 2 dout pin
const int HX711_sck_4 = 11; //mcu > HX711 no 2 sck pin

// Set up LCD address to 0x27 for a 16 chars and 2 line display
LiquidCrystal_I2C lcd(0x27, 16, 2);

//HX711 constructor (dout pin, sck pin)
HX711_ADC LoadCell_1(HX711_dout_1, HX711_sck_1); //HX711 1
HX711_ADC LoadCell_2(HX711_dout_2, HX711_sck_2); //HX711 2
HX711_ADC LoadCell_3(HX711_dout_3, HX711_sck_3); //HX711 3
HX711_ADC LoadCell_4(HX711_dout_4, HX711_sck_4); //HX711 4

// EEPROM address 0 used for storing calibration check byte
const int calVal_eepromAdress_1 = 4; // eeprom address for calibration value load cell 1 (4 bytes)
const int calVal_eepromAdress_2 = 8; // eeprom address for calibration value load cell 2 (4 bytes)
const int calVal_eepromAdress_3 = 12; // eeprom address for calibration value load cell 3 (4 bytes)
const int calVal_eepromAdress_4 = 16; // eeprom address for calibration value load cell 4 (4 bytes)

unsigned long t = 0;      // time counter
float d1 = 50;            // offset from wing LE to LoadCell 1
float d2 = 230;           // offset from wing LE to LoadCell 2
float a;                  // LoadCell 1 weight
float b;                  // LoadCell 2 weight
float c;                  // LoadCell 3 weight
float d;                  // Loadcell 4 weight
float w;                  // Total weight
float cog;                // CoG offset from Ref. point
#define calDone 0xAA      // Calibration check code
byte calStatus;           // saved Calibration code 
bool calFlag;             // true if calibration previously run

void setup()
{
  Wire.begin();
  lcd.begin();
  lcd.backlight();  // Turn on the LED backlight
  lcd.setCursor(2, 0);
  lcd.print("Starting...");
  Serial.begin(57600); delay(10);
  Serial.println();
  Serial.println("Starting...");

  float calibrationValue_1; // calibration value load cell 1
  float calibrationValue_2; // calibration value load cell 2
  float calibrationValue_3; // calibration value load cell 3
  float calibrationValue_4; // calibration value load cell 4

  EEPROM.get(0, calStatus); // Fetch value from to check if previously calibrated
  if(calStatus == calDone) calFlag = true;  // match = calibrated
  else calFlag = false;
  
  if(calFlag)
  {
    EEPROM.get(calVal_eepromAdress_1, calibrationValue_1); // fetch the value from eeprom
    EEPROM.get(calVal_eepromAdress_2, calibrationValue_2); // fetch the value from eeprom
    EEPROM.get(calVal_eepromAdress_3, calibrationValue_3); // fetch the value from eeprom
    EEPROM.get(calVal_eepromAdress_4, calibrationValue_4); // fetch the value from eeprom
  }
  else
  {
    calibrationValue_1 = 1; // Use default values if not yet calibrated
    calibrationValue_2 = 1;
    calibrationValue_3 = 1;
    calibrationValue_4 = 1;
  }

  LoadCell_1.begin();
  LoadCell_2.begin();
  LoadCell_3.begin();
  LoadCell_4.begin();
  unsigned long stabilizingtime = 2000; // tare preciscion can be improved by adding a few seconds of stabilizing time
  boolean _tare = true; //set this to false if you don't want tare to be performed in the next step
  byte loadcell_1_rdy = 0;
  byte loadcell_2_rdy = 0;
  byte loadcell_3_rdy = 0;
  byte loadcell_4_rdy = 0;
  
  while ((loadcell_1_rdy + loadcell_2_rdy + loadcell_3_rdy + loadcell_4_rdy) < 4) { //run startup, stabilization and tare, all modules simultaniously
    if (!loadcell_1_rdy) loadcell_1_rdy = LoadCell_1.startMultiple(stabilizingtime, _tare);
    if (!loadcell_2_rdy) loadcell_2_rdy = LoadCell_2.startMultiple(stabilizingtime, _tare);
    if (!loadcell_3_rdy) loadcell_3_rdy = LoadCell_3.startMultiple(stabilizingtime, _tare);
    if (!loadcell_4_rdy) loadcell_4_rdy = LoadCell_4.startMultiple(stabilizingtime, _tare);
  }
  if (LoadCell_1.getTareTimeoutFlag())
  {
    Serial.println("Timeout, check MCU>HX711 no.1 wiring and pin designations");
  }
  if (LoadCell_2.getTareTimeoutFlag())
  {
    Serial.println("Timeout, check MCU>HX711 no.2 wiring and pin designations");
  }
  if (LoadCell_3.getTareTimeoutFlag())
  {
    Serial.println("Timeout, check MCU>HX711 no.3 wiring and pin designations");
  }
  if (LoadCell_4.getTareTimeoutFlag())
  {
    Serial.println("Timeout, check MCU>HX711 no.4 wiring and pin designations");
  }
   
  LoadCell_1.setCalFactor(calibrationValue_1); // set calibration value (float)
  LoadCell_2.setCalFactor(calibrationValue_2); // set calibration value (float)
  LoadCell_3.setCalFactor(calibrationValue_3); // set calibration value (float)
  LoadCell_4.setCalFactor(calibrationValue_4); // set calibration value (float)
  
  Serial.println("Startup is complete");
  lcd.setCursor(1, 0);
  lcd.print("Startup done");
  delay(500);

  if(!calFlag) calibrate(); // if not yet calibrated, run calibration procedure
}

void loop()
{
  static boolean newDataReady = 0;
  const int serialPrintInterval = 500; //increase value to slow down serial print/display activity

  // check for new data/start next conversion:
  if (LoadCell_1.update()) newDataReady = true;
  LoadCell_2.update();
  LoadCell_3.update();
  LoadCell_4.update();

  //get smoothed value from data set
  if ((newDataReady))
  {
    if (millis() > t + serialPrintInterval)
    {
      a = LoadCell_1.getData();
      b = LoadCell_2.getData();
      c = LoadCell_3.getData();
      d = LoadCell_4.getData();
      Serial.print("Load_cell 1 output val: ");
      Serial.print(a);
      Serial.print("    Load_cell 2 output val: ");
      Serial.println(b);
      Serial.print("Load_cell 3 output val: ");
      Serial.print(c);
      Serial.print("    Load_cell 4 output val: ");
      Serial.println(d);
      
      newDataReady = 0;
      t = millis();

      w = a + b + c + d;  // total weight = sum of load cell values
      cog = ((d1*(a + c)) + (d2*(b + d)))/w;  // cog=weight dist/weight
      Serial.print("CoG distance from LE: ");
      Serial.print(cog);
      Serial.print("mm");
      Serial.print("   Total weight: ");
      Serial.print(w);
      Serial.println("g");

      lcd.setCursor(0, 0);   // Home lcd cursor
      lcd.print("Weight:         ");
      lcd.setCursor(8, 0);     // Start of weight value
      lcd.print(w, 0);
      lcd.print("g");
      lcd.setCursor(0, 1);
      lcd.print("   CoG:         ");
      lcd.setCursor(8, 1);
      if(w > 5)
      {
      lcd.print(cog, 0);
      }
      else lcd.print("---");
    }

  // receive command from serial terminal, send 't' to initiate tare operation:
  if (Serial.available() > 0)
  {
    char inByte = Serial.read();
    if (inByte == 't')
    {
      LoadCell_1.tareNoDelay();
      LoadCell_2.tareNoDelay();
      LoadCell_3.tareNoDelay();
      LoadCell_4.tareNoDelay();
    }
    else if(inByte == 'r') calibrate(); // send 'r' to Recalibrate all load cells
  }

  //check if last tare operation is complete
  if (LoadCell_1.getTareStatus() == true)
  {
    Serial.println("Tare load cell 1 complete");
  }
  if (LoadCell_2.getTareStatus() == true)
  {
    Serial.println("Tare load cell 2 complete");
  }
  if (LoadCell_3.getTareStatus() == true)
  {
    Serial.println("Tare load cell 3 complete");
  }
  if (LoadCell_4.getTareStatus() == true)
  {
    Serial.println("Tare load cell 4 complete");
  }
 }
}

void calibrate()
{
  Serial.println("***");
  Serial.println("Start calibration:");
  Serial.println("Place all load cells on a level stable surface.");
  Serial.println("Remove any load applied to the load cells.");
  Serial.println("Send 't' from serial monitor to set the tare offset.");

  lcd.setCursor(0, 0);   // Home lcd cursor
  lcd.print(" Calibrating via");
  lcd.setCursor(0, 1);
  lcd.print("    Serial      ");

// Load cell 1
  
  bool _resume = false;
  while (_resume == false)
  {
    LoadCell_1.update();
    if(Serial.available() > 0)
    {
       byte inByte = Serial.read();
       if (inByte == 't')
       {
        LoadCell_1.tareNoDelay();
        if (LoadCell_1.getTareStatus() == true) Serial.println("Load Cell 1 Tare complete");
        _resume = true;
       }
    }   
  }

  Serial.println("Now, place your known mass on loadcell 1.");
  Serial.println("Then send the weight of this mass (i.e. 100.0) from serial monitor.");

  float known_mass = 0;
  _resume = false;
  while (_resume == false)
  {
    LoadCell_1.update();
    if (Serial.available() > 0)
    {
      known_mass = Serial.parseFloat();
      if (known_mass != 0)
      {
        Serial.print("Known mass is: ");
        Serial.println(known_mass);
        _resume = true;
      }
    }
  }

  LoadCell_1.refreshDataSet(); //refresh the dataset to be sure that the known mass is measured correct
  float newCalibrationValue = LoadCell_1.getNewCalibration(known_mass); //get the new calibration value

  Serial.print("New calibration value has been set to: ");
  Serial.print(newCalibrationValue);
  Serial.println(", saving this in EEPROM as Cell 1 calibration value (calFactor).");

  _resume = false;
  while (_resume == false)
  {
    EEPROM.put(calVal_eepromAdress_1, newCalibrationValue);
    EEPROM.get(calVal_eepromAdress_1, newCalibrationValue);
    Serial.print("Value ");
    Serial.print(newCalibrationValue);
    Serial.print(" saved to EEPROM address: ");
    Serial.println(calVal_eepromAdress_1);
    Serial.println("");
    _resume = true;
  }

// Load cell 2

  Serial.println(""); // space between text blocks
  Serial.println("");
  Serial.println("Remove any load applied to the load cells.");
  Serial.println("Send 't' from serial monitor to set the tare offset.");
  
  _resume = false;
  while (_resume == false)
  {
    LoadCell_2.update();
    if(Serial.available() > 0)
    {
       byte inByte = Serial.read();
       if (inByte == 't')
       {
        LoadCell_2.tareNoDelay();
        if (LoadCell_2.getTareStatus() == true) Serial.println("Load Cell 2 Tare complete");
        _resume = true;
       }
    }   
  }

  Serial.println("Now, place your known mass on loadcell 2.");
  Serial.println("Then send the weight of this mass (i.e. 100.0) from serial monitor.");

  known_mass = 0;
  _resume = false;
  while (_resume == false)
  {
    LoadCell_2.update();
    if (Serial.available() > 0)
    {
      known_mass = Serial.parseFloat();
      if (known_mass != 0)
      {
        Serial.print("Known mass is: ");
        Serial.println(known_mass);
        _resume = true;
      }
    }
  }

  LoadCell_2.refreshDataSet(); //refresh the dataset to be sure that the known mass is measured correct
  newCalibrationValue = LoadCell_2.getNewCalibration(known_mass); //get the new calibration value

  Serial.print("New calibration value has been set to: ");
  Serial.print(newCalibrationValue);
  Serial.println(", saving this in EEPROM as Cell 2 calibration value (calFactor).");

  _resume = false;
  while (_resume == false)
  {
    EEPROM.put(calVal_eepromAdress_2, newCalibrationValue);
    EEPROM.get(calVal_eepromAdress_2, newCalibrationValue);
    Serial.print("Value ");
    Serial.print(newCalibrationValue);
    Serial.print(" saved to EEPROM address: ");
    Serial.println(calVal_eepromAdress_2);
    Serial.println("");
    _resume = true;
   }

// Load Cell 3 **********************

  Serial.println(""); // space between text blocks
  Serial.println("");
  Serial.println("Remove any load applied to the load cells.");
  Serial.println("Send 't' from serial monitor to set the tare offset.");
  
  _resume = false;
  while (_resume == false)
  {
    LoadCell_3.update();
    if(Serial.available() > 0)
    {
       byte inByte = Serial.read();
       if (inByte == 't')
       {
        LoadCell_3.tareNoDelay();
        if (LoadCell_3.getTareStatus() == true) Serial.println("Load Cell 3 Tare complete");
        _resume = true;
       }
    }   
  }

  Serial.println("Now, place your known mass on loadcell 3.");
  Serial.println("Then send the weight of this mass (i.e. 100.0) from serial monitor.");

  known_mass = 0;
  _resume = false;
  while (_resume == false)
  {
    LoadCell_3.update();
    if (Serial.available() > 0)
    {
      known_mass = Serial.parseFloat();
      if (known_mass != 0)
      {
        Serial.print("Known mass is: ");
        Serial.println(known_mass);
        _resume = true;
      }
    }
  }

  LoadCell_3.refreshDataSet(); //refresh the dataset to be sure that the known mass is measured correct
  newCalibrationValue = LoadCell_3.getNewCalibration(known_mass); //get the new calibration value

  Serial.print("New calibration value has been set to: ");
  Serial.print(newCalibrationValue);
  Serial.println(", saving this in EEPROM as Cell 3 calibration value (calFactor).");

  _resume = false;
  while (_resume == false)
  {
    EEPROM.put(calVal_eepromAdress_3, newCalibrationValue);
    EEPROM.get(calVal_eepromAdress_3, newCalibrationValue);
    Serial.print("Value ");
    Serial.print(newCalibrationValue);
    Serial.print(" saved to EEPROM address: ");
    Serial.println(calVal_eepromAdress_3);
    Serial.print("");
    _resume = true;
  }

 // Load Cell 4 **************************

  Serial.println(""); // space between text blocks
  Serial.println("");
  Serial.println("Remove any load applied to the load cells.");
  Serial.println("Send 't' from serial monitor to set the tare offset.");
  
  _resume = false;
  while (_resume == false)
  {
    LoadCell_4.update();
    if(Serial.available() > 0)
    {
       byte inByte = Serial.read();
       if (inByte == 't')
       {
        LoadCell_4.tareNoDelay();
        if (LoadCell_4.getTareStatus() == true) Serial.println("Load Cell 4 Tare complete");
        _resume = true;
       }
    }   
  }

  Serial.println("Now, place your known mass on loadcell 4.");
  Serial.println("Then send the weight of this mass (i.e. 100.0) from serial monitor.");

  known_mass = 0;
  _resume = false;
  while (_resume == false)
  {
    LoadCell_4.update();
    if (Serial.available() > 0)
    {
      known_mass = Serial.parseFloat();
      if (known_mass != 0)
      {
        Serial.print("Known mass is: ");
        Serial.println(known_mass);
        _resume = true;
      }
    }
  }

  LoadCell_4.refreshDataSet(); //refresh the dataset to be sure that the known mass is measured correct
  newCalibrationValue = LoadCell_4.getNewCalibration(known_mass); //get the new calibration value

  Serial.print("New calibration value has been set to: ");
  Serial.print(newCalibrationValue);
  Serial.println(", saving this in EEPROM as Cell 4 calibration value (calFactor).");

  _resume = false;
  while (_resume == false)
  {
    EEPROM.put(calVal_eepromAdress_4, newCalibrationValue);
    EEPROM.get(calVal_eepromAdress_4, newCalibrationValue);
    Serial.print("Value ");
    Serial.print(newCalibrationValue);
    Serial.print(" saved to EEPROM address: ");
    Serial.println(calVal_eepromAdress_4);
    Serial.print("");
    _resume = true;
  }

  calStatus = calDone; // set calStatus to indicate calibration done
  EEPROM.put(0, calStatus);  // Save calibration status to EEPROM
  calFlag = true; // set flag to confirm calibration has been done
  
  Serial.println("End calibration");
  Serial.println("***");
  Serial.println("To re-calibrate, send 'r' from serial monitor.");
}
